#!/usr/bin/env python
# -*- coding: utf-8 -*-

##
#
# Name: 	mina.py
# Author: 	Lenka Turonova
# Date: 	14. 4. 2015
# Version: 	7.0
# Description:
# Program load data from file and run program
# implemented acording to improved Raskin 
# algorithm.
# 
##

## LIBRARIES ##
import sys
import copy
import re
import time
import string
import time
from random import random,randrange,randint

# start timer
start_time = time.time()

##-------------------------------------------------------------------##
## CONSTANTS ##
NONE = -1
VARS = 0
RULES = 1
INIT = 2
TARGET = 3
INVAR = 4
##-------------------------------------------------------------------##
## GLOBAL VARIABLES ##
bad = NONE
badState = NONE
variables = []
transition = []
invariants = []
x0 = NONE

dBad = [[bad]] 	# backward search
P = []			# forward search
flag = True		# flag for invariant violence
it = 0			# number of iteration CEGAR
path = {}		# forward
pathSeq = []	# forward path
tree = []		# structure for creating tree structure
##-------------------------------------------------------------------##

## INIT LIST REPRESENTING STATE ##
def initState():
	state = {}
	for item in variables:
		state[item] = 0.0
	return state
##-------------------------------------------------------------------##

## LOAD PARAMETERS FROM INPUT FILE ##
def loadFile():
	global bad
	global x0
	global variables
	global transition
	global badState
	global invariants
	init = []
	target = []
	flag = NONE
	invariant = []
	rules = []
	targ = []
	rText = ''
	# read parameters
	if len(sys.argv) > 1:
		name = sys.argv[1]
	else:
		print("Wrong name or path to file!")
		sys.exit(0)
	# open file
	f=file(name,'r') 

	# read file
	for r in f:  
		if flag == VARS: # load variables
			m = re.compile("\s+").split(r)
			if m is not None:
				variables = m[1:-1]
				flag = NONE
			else:
				print "Error"
		if flag == INIT or flag == TARGET or flag == INVAR:
			m = re.compile(",").split(r)
			if m is not None:
				for item in m:
					n = re.match(r'[\s]*([A-Za-z0-9_]+)[\s]*([=><]+)[\s]*([0-9]+)', item)
					if n is not None:
						if flag == INIT:	# load initial state
							init += [[n.group(1), n.group(2), n.group(3)]]
						if flag == TARGET:	# load targets
							targ += [[n.group(1), n.group(2), n.group(3)]]
						if flag == INVAR:	# load invariants
							invariant += [[n.group(1), n.group(2), n.group(3)]]
				if flag == TARGET and targ != []:
					target += [targ]
					targ = []
				if flag == INVAR:
					invariants += [invariant]
					invariant = []
			else:
				print "Error"
		if flag == RULES:	# load rules
			if '#' not in r and not r.isalpha() and not r.isspace():
				rText += r.lstrip().rstrip()
		m = re.match(r'vars', r)
		if m is not None:
			flag = VARS
		m = re.match(r'rules', r)
		if m is not None:
			flag = RULES
		m = re.match(r'init', r)
		if m is not None:
			flag = INIT
		m = re.match(r'target', r)
		if m is not None:
			flag = TARGET
		m = re.match(r'invariants', r)
		if m is not None:
			flag = INVAR
	f.close()

	# load rules
	rl = []
	rt = re.compile(";").split(rText)
	for item in rt:
		if "->" in item:
			rl += [re.compile("->").split(item)]
	rul = []
	for item in rl:
		set = [re.compile(",").split(item[0])] + [re.compile(",").split(item[1])]
		rul += [set]
	right = []
	left = []
	for item in rul:
		for r in item[0]:
			n = re.match(r'[\s]*([A-Za-z0-9_]+)[\s]*([=><]+)[\s]*([0-9]+)', r)
			if n is not None:
				right += [[n.group(1), n.group(2), n.group(3)]]
		for r in item[1]:
			n = re.match(r'[\s]*([A-Za-z0-9_\']+)[\s]*=[\s]*([A-Za-z0-9_]+)[\s]*([+-])[\s]*([0-9]+)[\s]*', r)
			if n is not None:
				left += [[n.group(1), n.group(2), n.group(3), n.group(4)]]
		rules += [[right]+[left]]
		right = []
		left = []

	## INITIALIZATION OF INITIAL STATE ##
	x0 =  initState()
	for item in init:
		x0[item[0]] = float((item[2]))

	## INITIALIZATION OF BAD STATE ##
	badState =  initState()
	bad = []
	for item in target:
		for state in item:
			badState[state[0]] = float((state[2]))
		bad += [badState]
		badState =  initState()
	transition = []
	plus = []
	minus = []
	countR = 0

	## INITIALIZATION OF RULES ##
	for item in rules:
		countR += 1
		for eq in item[1]:
			if(eq[2] == "+"):
				if(float(eq[3]) > 1.0):
					plus += [[eq[1], int(eq[3])]] 
				else:
					plus += [[eq[1]]]
			if(eq[2] == "-"):
				if(float(eq[3]) > 1.0):
					minus += [[eq[1], int(eq[3])]] 
				else:
					minus += [[eq[1]]]
		for eq in item[0]:
			flag = False
			for ep in item[1]:
				if(eq[0] in ep):
					flag = True
					break
			if not flag:
				minus += [[eq[0]]]
				plus += [[eq[0]]]
		name = "t" + str(countR)
		transition += [[name, minus, plus]]
		plus = []
		minus = []
##-------------------------------------------------------------------##

## PRINT MARKING ##
def printlabeling(label):
	string = ""
	for index in label:
			if label[index] > 0.0:
				string += str(index) + ":" + str(label[index]) + " "
	print string
	return

## PRINT LOADED PARAMETERS ##
def printParameters():
	global x0
	global transition
	global variables
	global bad
	# print loaded parameters
	print("--------------------------------------------------------")
	print "PARAMETERS"
	print("--------------------------------------------------------\n")
	print "variables"
	for item in variables:
		print(item)
	print "\nx0"
	printlabeling(x0)
	print "\nbad"
	for item in bad:
		printlabeling(item)
	print "\ntransitions"
	for item in transition:
		print item
	print("--------------------------------------------------------")
	print("--------------------------------------------------------")
##-------------------------------------------------------------------##

## MARKING TO STRING ##
def markingToString(label):
	count = 0
	name = ''
	for index in label.keys():
		count = label[index]
		while(count > 0.0):
			name += str(index)
			name += "_"
			count -= 1.0
	return name[:-1]
##-------------------------------------------------------------------##

def LabelToString(label):
	count = 0
	name = ''
	for index in label.keys():
		count = label[index]
		while(count > 0.0):
			name += str(index)
			name += "_"
			count -= 1.0
	return name[:-1]

##  CREATE FILE WITH GRAPH  ##
def createFile():
	global tree
	global it 
	
	# create graph description for .dot file
	text = "digraph G"+ str(it)+ " {\n"
	for item in tree:
		text += item[2] +  "[shape=rect,fixedsize=true,height=0.3,width=0.8,style=\"rounded\"];\n"
		text += LabelToString(item[0]) + "->" 
		text += item[2] + "->" 
		text += LabelToString(item[1]) + ";\n"
	text += "}\n"
	# write into file
	file = open('output/graph'+str(it)+'.dot', 'w+') # 'at' overwrite file
	file.write(text)
	file.close()
##-------------------------------------------------------------------##

## PETRI NET CLASS ##
class PetriNetBase:
    # constructor
    def __init__(this, names, transitionSpecs):
        this.names = names
        this.transitions = this.BuildTransitions(transitionSpecs)

    def BuildTransitions(this, transitionSpecs):
        transitions = []
        for (transitionName, inputSpecs, outputSpecs) in transitionSpecs:
            transition = Transition(transitionName)

            for degreeSpec in inputSpecs:
                this.SetDegree(transition.inputMap, degreeSpec)

            for degreeSpec in outputSpecs:
                this.SetDegree(transition.outputMap, degreeSpec)

            transitions.append(transition)

        return transitions 

    def SetDegree(this, dictionary, degreeSpec):
        name = degreeSpec[0]

        if len(degreeSpec) == 2:
            degree = degreeSpec[1]
        else: 
            degree = 1

        dictionary[name] = degree
##-------------------------------------------------------------------##

class Transition:
    def __init__(this, transitionName):
       this.transitionName = transitionName
       this.inputMap = {} 
       this.outputMap = {} 

    ## FIRE STEP BACK ##
    def FireBack(this, labeling, old):
       new = initState()

       for item in new.keys():
          new[item] = labeling[item]
       for outputName in this.outputMap.keys():
          new[outputName] = new[outputName] - this.outputMap[outputName] 
          if new[outputName] < 0.0:
				new[outputName] = 0.0
       for inputName in this.inputMap.keys():
          new[inputName] = new[inputName] + this.inputMap[inputName] 

       return new

	## CHECK TRANSITION IS ENABLED ##
    def IsEnabledBack(this, labeling):
       for output in this.outputMap.keys():
          if labeling[output] >= this.outputMap[output]:
              return True # good to go 
       return False  # not enough tokens     

	## PRINT TRANSITION NAME ##
    def PrintName(this):
		#print this.transitionName
		return this.transitionName
		
##-------------------------------------------------------------------##

class PetriNet(PetriNetBase):

	## CONVERT STRING TO labeling ##
    def StringToLabel(this, name):
		label = initState()
		for index in label.keys():
			label[index] = float(name.count(index))
		return label

	## CONVERT labeling TO STRING ##
    def LabelToString(this, label):
		count = 0
		name = ''
		for index in label.keys():
			count = label[index]
			while(count > 0.0):
				name += str(index)
				name += "_"
				count -= 1.0
		return name[:-1]

	## CONVERT labeling TO LIST ##
    def LabelToList(this, label):
		count = 0
		set = []
		for index in label.keys():
			count = label[index]
			while(count > 0.0):
				set += [str(index)]
				count -= 1.0
		return set
				

	## ADD ITEMS TO Pi ##
    def Forward(this, pSet, dSet):
		global x0
		global path
		global pathSeq
		
		pathSeq = []
		flag = True
		flagA = True

		# P0 enumeration
		this.invariant = copy.deepcopy(x0)
		for item in dSet:
			if(this.CheckInvariant(item)):
				pSet += [item]	# add items that are a part of initial state
	
		while(flagA): 		# forward search step by step
			flagA = False 	# new element to pSet not added
			# complement recalculation
			Comp = [] 		
			for item in dSet:	
				if item not in pSet:
					Comp += [item]
			for element in dSet:	# states in D\P
				if element not in pSet:
					this.labeling = element
					items = filter(lambda transition: transition.IsEnabledBack(this.labeling), this.transitions) # one step back
					for index in range(len(items)):	
						newItem = items[index].FireBack(this.labeling, element) # each predecessor
						for item in Comp: # states in D\P
							this.invariant = newItem
							if(this.CheckInvariant(item)):	# check if state is valid
								flag = False
								break
						if flag and this.labeling not in pSet:
							pSet += [this.labeling] 
							#printlabeling(this.labeling)
							path[this.LabelToString(this.labeling)] = items[index].PrintName()
							pathSeq += [items[index]]
							flagA = True # new element to pSet added
						flag = True
		return pSet

	
	## GENERATE SUBWORDS ##
    def Generate(this, position, start, N, i, singles, subset, all):
		if position < i:
			for state in range(start, N):   
				subset[position]=singles[state]        
				this.Generate(position+1,state+1, N, i, singles, subset, all)      
		else:
			set = ''
			for item in subset:
				set += item
			all += [this.StringToLabel(set)]
		  
	## POSSIBLE SUBWORDS ##
    def AllSubwords(this, state):
		set = this.LabelToList(state)
		singles = []
		all = []

		for i in range(len(set)):
			all += [this.StringToLabel(set[i])]  # subwords of length 1
			singles += [set[i]]

		N = len(singles)
		for i in range(2, len(singles)+1):
			subset = [0] * i
			this.Generate(0, 0, N, i, singles, subset, all)     
		return all

			  
	## POSSIBLE COMBINATIONS OF SUBWORD ##
    def AllCombination(this, sets):
		set = []
		for item in sets:
			set +=[this.LabelToString(item)]
		singles = []
		all = []

		for i in range(len(set)):
			all += [this.StringToLabel(set[i])]  # subwords of length 1
			singles += [set[i]]

		N = len(singles)
		for i in range(2, len(singles)+1):
			subset = [0] * i
			this.Generate(0, 0, N, i, singles, subset, all)     
		return all
								
	
	## SUBWORD WITHIN P ##
    def Subwords(this, state, pSet):
		set = []
		for item in pSet:
			this.invariant = state
			if(this.CheckInvariant(item)):	# check all subwords of predecessor  #i v cx;kzcf;i cck   x		
				set += [item]
		return set

	## CREATE TREE USING BACKWARD SEARCH ##
    def BackwardTree(this, pSet, dSet, badState):
		global path
		global x0
		global tree
		global it
		newSet = []
		index = 0
		initAbs = []
		treeStr = [[badState]]
		treePre = [badState]
		set = [this.LabelToString(badState)]
		pred = []
		tree = []
		
		this.labeling = badState
		## intial abstraction
		for item in pSet:
			this.invariant = copy.deepcopy(x0)
			if(this.CheckInvariant(item)):	# check all subwords of predecessor
				initAbs += [item]

		## intial abstraction of bad state	
		for item in pSet:
			this.invariant = badState
			if(this.CheckInvariant(item)):	# check all subwords of predecessor
				this.labeling = item
				for i in range(len(this.transitions)):	# find transaction of path from abstraction of bad state
					if(path[this.LabelToString(item)] == this.transitions[i].transitionName):
						break
				pred = this.transitions[i].FireBack(this.labeling, item)	# save predecessor
				treeStr += [this.Subwords(pred, pSet)]	# save all subword of predecessor
				treePre += [pred]			# save predecessor
				set += [this.LabelToString(pred)]	# save predecessor as string
				tree += [[badState, pred, this.transitions[i].transitionName]]	# create new note in tree


		## create tree
		while(index+1 != len(treeStr)):
			index += 1
			for item in treeStr[index]: # for all predecessors
				if item not in initAbs: # state is not initial abstraction
					this.labeling = item				
					for i in range(len(this.transitions)):	# search for transaction back
						if(path[this.LabelToString(item)] == this.transitions[i].transitionName):
							break
					pred = this.transitions[i].FireBack(this.labeling, item)	# fire transaction
					if this.LabelToString(pred) not in set:	# if predecessor not inspected
						treePre += [pred]	# save predecessors
						set += [this.LabelToString(pred)]	# save predecessor as string
						treeStr += [this.Subwords(pred, pSet)]	# add all subwords
						tree += [[treePre[index], pred, this.transitions[i].transitionName]]	# create new note in tree (whole state - predecessor)

		## add into set
		for item in treePre:
			for subword in this.AllSubwords(item):
				if subword not in dSet: # all subwords not in D  
					newSet += [subword]
		return newSet

    def EnabledTransitionsBack(this):
        return filter(
            lambda transition: transition.IsEnabledBack(this.labeling),
            this.transitions)

    def FireOneRuleBack(this):
        this.SelectRandom(this.EnabledTransitionsBack()).FireBack(this.labeling)

    def IsHalted(this):
		return len(this.EnabledTransitionsBack()) == 0

    def SelectRandom(this, items):
        randomIndex = randrange(len(items))
        return items[randomIndex]

	## CHECK INVARIANT ##
    def CheckInvariant(this, label):
		for index in label:
			if label[index] > this.invariant[index]: 
				#print index
				return False # invariant invalid
		return True  # invariant valid

    def printlabeling(this, label):
		string = ""
		for index in label:
		    if label[index] > 0.0:
				string += str(index) + ":" + str(label[index]) + " "
		print string
		return
##-------------------------------------------------------------------##
   

## PRINT ABSTRACT MARKING ##
def printlabeling(dict):
	string = ""
	count = 0
	name = ''
	for index in dict.keys():
		count = dict[index]
		while(count > 0.0):
			name += str(index) + '_'
			count -= 1.0
	print name[:-1]
	return
##-------------------------------------------------------------------##

## LEN OF STATE ##
def lenState(label1, label2):
	count1 = 0
	count2 = 0
	for index in label1:
		if label1[index] > 0.0:
			count1 += label1[index]
		if label2[index] > 0.0:
			count2 += label2[index]
	return int(count2-count1)
##-------------------------------------------------------------------##


####################
#       MAIN       #
####################
# load input file
loadFile()	
# x0  			initial state
# bad 			a set of incorrect states
# transition 	a set of transitions
# variables 	a set of states 
printParameters()               

## PETRI NET INITIATION ##
petriNet = PetriNet(variables, transition)
D = []
for item in bad:
	D+=[item]
	
lastD = []
initD = []
for item in D:
	lastD += [item]
	initD += [item]

starttime = time.time() 			# set time
## forward search
it += 1
P = []
path = {}
flag = False
P = petriNet.Forward(P, D)

for item in bad:
	if item in P: # find bad state in P
		flag = True
		badState = item # save bad state for backward run

		
print '------------------------'
while(flag):
	flag = False
	lenSet = []
	newSet = petriNet.BackwardTree(P, D, badState)
	createFile()
	if(len(newSet) == 0):
		print "Tree is empty."
		for item in newSet:
			printlabeling(item)
		break
	sorted(newSet, cmp=lenState)
	D += [newSet[0]]
	P = []
	path = {}
	P = petriNet.Forward(P, D)
	
	for item in bad:
		if item in P: # find bad state in P
			flag = True
			badState = item # save bad state for backward run
			break
	it += 1
	

print '---------------'
print 'Elements of D'
for item in D:
	printlabeling(item)


print "\nNumber of iterations: " + str(it)		
print 'Items in D: ' + str(len(D))
print 'Time: ' + str(round(time.time() - starttime, 3)) + 's'

	